from io import StringIO
from typing import Dict, List

import pytest

from boiler.definitions import ActorType
from boiler.models import Box, Detection
from boiler.serialization import ParseError
from boiler.serialization.kw18 import KW18FileSet


@pytest.fixture
def kw18():
    kw18_file = StringIO(
        """#
1 1037 4800 0 0 0 0 304 525 288 489 320 561 -1 444 444 0 2400 -1
1 1037 5836 0 0 0 0 1895 357.5 1881 331 1909 384 -1 444 444 0 2918 -1
2 94 5593 0 0 0 0 1325.5 1059.5 1216 1048 1435 1071 -1 444 444 0 2796.5 -1
2 94 5686 0 0 0 0 1345.5 1050.5 1235 1030 1456 1071 -1 444 444 0 2843 -1
3 75 5712 0 0 0 0 1360.5 1051.5 1234 1031 1487 1072 -1 444 444 0 2856 -1
3 75 5786 0 0 0 0 1216.5 1064.5 1090 1058 1343 1071 -1 444 444 0 2893 -1
"""
    )
    types_file = StringIO(
        """#
1 Person
2 Vehicle
3 Vehicle
"""
    )
    txt_file = StringIO(
        """#
0 8 144 -1.00000000000000005250e+300 4800 -1.00000000000000005250e+300 5836 1.0000000000e+000 1.000000e+000 1.000000e+000 0.000000e+000 0.000000e+000 1 1
0 10 145 -1.00000000000000005250e+300 5593 -1.00000000000000005250e+300 5686 1.0000000000e+000 1.000000e+000 1.000000e+000 0.000000e+000 0.000000e+000 1 2
0 44 146 -1.00000000000000005250e+300 5712 -1.00000000000000005250e+300 5786 1.0000000000e+000 1.000000e+000 1.000000e+000 0.000000e+000 0.000000e+000 1 3
"""  # noqa
    )
    yield {
        'kw18': kw18_file,
        'regions': None,
        'types': types_file,
        'txt': txt_file,
    }


@pytest.fixture
def geom_input():
    return StringIO(
        (
            '210 85 6813 0 0 0 DEADBEEF FEEDF00D 0 457 569 478 612\n'
            '210 85 6871 0 0 0 0 0 0 414 581 435 624\n'
            '210 85 6898 0 0 0 0 0 0 404 591 425 634\n'
            '211 107 7152 0 0 0 x 0 0 461 571 482 614\n'
            '211 107 7175 0 0 nonsense x 0 0 440 572 461 615\n'
            '211 107 7259 0 0 x x 0 0 402 592 423 635\n'
        )
    )


@pytest.fixture
def comment_input():
    return StringIO('# comment row \n')


def test_deserialize_types_ex1(comment_input):
    kw18 = KW18FileSet(types=comment_input)
    actor_map = kw18.read_actor_types()
    assert len(actor_map.values()) == 0


def test_deserialize_types_ex2():
    sinput = StringIO('1 Vehicle\n2 Vehicle\n1 Person\n')
    kw18 = KW18FileSet(types=sinput)
    actor_map = kw18.read_actor_types()
    assert len(actor_map.values()) == 2
    assert 1 in actor_map and 2 in actor_map
    assert 0 not in actor_map
    assert actor_map[1] == ActorType.PERSON
    assert actor_map[2] == ActorType.VEHICLE


def test_deserialize_geom_ex2(geom_input):
    kw18 = KW18FileSet(types=StringIO(''), kw18=geom_input)
    actor_map = kw18.read_detections()
    assert len(actor_map.values()) == 2
    assert 210 in actor_map and 211 in actor_map
    assert len(actor_map[210]) == 3
    assert len(actor_map[211]) == 3
    assert actor_map[210][0].frame == 6813
    assert actor_map[210][0].keyframe is True
    assert actor_map[210][0].box == Box(left=457, top=569, right=478, bottom=612)


def test_malformed_activities():
    actor_type_map = {0: ActorType.PERSON}
    detection_map: Dict[int, List[Detection]] = {0: []}
    kw18 = KW18FileSet()
    with pytest.raises(ParseError) as err:
        kw18._parse_activity_row(
            '0 24 144 0 4238 4312 0 0 0 0 0 1 216'.split(' '), actor_type_map, detection_map
        )
    assert 'Could not parse activity from the *.txt file' in str(err)


def test_deserialize_activities_ex1(kw18):
    kw18 = KW18FileSet(**kw18)
    activity_map = kw18.read().activity_map
    assert len(activity_map.values()) == 3
    assert len(activity_map[144].actors) == 1
    assert activity_map[144].begin == 4800
    assert activity_map[144].end == 5836
    assert activity_map[145].begin == 5593
    assert activity_map[145].end == 5686


# TODO: test multi-actor activity


def test_deserialize_regions(geom_input):
    sinput = StringIO(
        (
            '# [0:actor_id] [1:frame] [2:is_keyframe] [?] ...\n'
            '0 0 1 nonsense\n'
            '210 6813 0\n'
            '210 6898 400\n'
        )
    )
    kw18 = KW18FileSet(regions=sinput)
    interpolated_frames = kw18.read_interpolated_frames()
    assert len(interpolated_frames.values()) == 1
    assert {210} == set(interpolated_frames.keys())
    assert interpolated_frames[210] == {6813}



"""
Refiner answer deserialization into boiler models.
"""
from itertools import groupby
import json
from typing import Any, Dict, List, Optional, Union

from intervals import IntInterval

from boiler import activity_migration, definitions, models
from boiler import BoilerError


def invert_type_mapping(map: Dict[str, int]) -> Dict[int, str]:
    """Invert a refiner type mapping."""
    # For some reason, refiner answers provide the inverse of what you actually
    # need....
    return {v: k for k, v in map.items()}


def is_keyframe(detection: Dict[str, Any]) -> bool:
    # TODO: do we want to handle other types?
    return detection['src'] in ('ground-truth', 'truth')


def deserialize_detection(detection: Dict[str, Any]) -> models.Detection:
    box = detection['bbox']
    # refiner doesn't reliably send bounding boxes coordinates in the correct order
    return models.Detection(
        frame=detection['frame'],
        box=models.Box(
            left=min(box[0], box[2]),
            top=min(box[1], box[3]),
            right=max(box[0] + 1, box[2]),  # also protect against zero area boxes
            bottom=max(box[1] + 1, box[3]),
        ),
        keyframe=is_keyframe(detection),
    )


def deserialize_answers(answer_string: Union[str, Dict[str, Any]]) -> List[models.Activity]:
    if isinstance(answer_string, str):
        answer = json.loads(answer_string)
    else:
        answer = answer_string

    if 'types' not in answer or 'activities' not in answer or 'detections' not in answer:
        raise BoilerError('answer did not contain all required keys')

    types = answer['types']
    activities: List[models.Activity] = []
    for activity_id, activity in enumerate(answer['activities']):
        flat_detections = sorted(answer['detections'], key=lambda d: d['id1'])

        actor_type_map: Dict[int, definitions.ActorType] = {}
        for actor_type in types:
            actor_type_string = list(actor_type['cset3'].keys())[0].lower()
            actor_type_map[actor_type['id1']] = definitions.ActorType.from_string(actor_type_string)

        actor_map = {actor['id1']: actor for actor in activity.get('actors', [])}
        detection_map = groupby(flat_detections, lambda d: d['id1'])
        actors: List[models.Actor] = []

        for actor_id, detection_iterator in detection_map:
            answer_actor = actor_map[actor_id]
            actor_type = actor_type_map[actor_id]
            time_range: Optional[IntInterval] = None
            if 'timespan' in answer_actor:
                timespan = answer_actor['timespan']

                # sometimes refiner returns a dict rather than a list for the
                # actor's timespan /shrug
                if isinstance(timespan, dict):
                    timespan = [timespan['0']]
                time_range = IntInterval(timespan[0]['tsr0'])

            detections = [deserialize_detection(d) for d in detection_iterator]

            # sometimes the actor is missing a time range entirely, so
            # we fall back to deriving it from the detections
            if time_range is None:
                time_range = IntInterval([detections[0].frame, detections[-1].frame])

            actor = models.Actor(
                actor_id=actor_id,
                begin=time_range.lower,
                end=time_range.upper,
                actor_type=actor_type,
                detections=detections,
            )
            actors.append(actor)

        activity_id = activity['id2']
        activity_type_map = invert_type_mapping(activity['act2'])
        activity_type_name = activity_type_map[activity_id]
        if activity_type_name in activity_migration.activity_type_mapping:
            activity_type = activity_migration.activity_type_mapping[activity_type_name]
        else:
            activity_type = definitions.ActivityType(activity_type_map[activity_id])
        frame_range = IntInterval(activity['timespan'][0]['tsr0'])

        activities.append(
            models.Activity(
                activity_id=activity_id,
                activity_type=activity_type,
                begin=frame_range.lower,
                end=frame_range.upper,
                actors=actors,
            )
        )
    return activities


def deserialize_answer(answer_string: Union[str, Dict[str, Any]]) -> models.Activity:
    activities = deserialize_answers(answer_string)
    if len(activities) != 1:
        raise BoilerError('answer should contain exactly 1 activity')
    return activities[0]



from math import inf
import re
from typing import List, Tuple

from boiler import BoilerError, BoilerWarning
from boiler.definitions import activity_spec, ActivityType, actor_codes
from boiler.models import Activity, ActivityList, Actor, Detection

# Ideally, we set this higher, but theoretically there could be a 3 frame track...
MAX_KEYFRAME_RATIO = 0.75

PROBLEMS = Tuple[List[BoilerWarning], List[BoilerError]]


def validate_activity_actor_types(
    actor_string: str, activity_type: ActivityType, activity_id: int
) -> PROBLEMS:
    warnings: List[BoilerWarning] = []
    errors: List[BoilerError] = []
    pattern = activity_spec.get(activity_type)
    if pattern is not None:
        sorted_actor_string = ''.join(sorted(actor_string))
        pattern = f'^{pattern}$'
        result = re.match(pattern, sorted_actor_string)
        if result is None:
            warnings.append(
                BoilerWarning(
                    f'activity_id={activity_id} activity spec validation failed for {activity_type.value}:'  # noqa:E501
                    f' {sorted_actor_string} found, {activity_spec[activity_type]} expected'
                )
            )

    return warnings, errors


def validate_detection(detection: Detection, activity_id: int, frame: int) -> PROBLEMS:
    """
    * detections have 4 int corners
    * detection: top < bottom, left < right
    """
    warnings: List[BoilerWarning] = []
    errors: List[BoilerError] = []
    prefix = f'activity_id={activity_id} detection_frame={frame}'
    box = detection.box
    if not (box.left < box.right):
        errors.append(BoilerError(f'{prefix} left ({box.left}) should be < right ({box.right})'))
    if not (box.top < box.bottom):
        errors.append(BoilerError(f'{prefix} top ({box.top}) should be < bottom ({box.bottom})'))
    return warnings, errors


def validate_actor(actor: Actor, activity_id: int, actor_id: int) -> PROBLEMS:
    warnings: List[BoilerWarning] = []
    errors: List[BoilerError] = []
    prefix = f'activity_id={activity_id} actor_id={actor_id}'
    keyframe_count = 0
    last = None
    for detection in actor.detections:
        if last and detection.frame <= last.frame:  # type: ignore
            errors.append(  # type: ignore
                BoilerError(
                    f'{prefix} detection_frame={detection.frame} was not greater than {last.frame}'
                )
            )
        last = detection
        detection_warnings, detection_errors = validate_detection(
            detection, activity_id, detection.frame
        )
        if detection.keyframe:
            keyframe_count += 1

    if not (len(actor.detections) >= 2):
        errors.append(BoilerError(f'{prefix} actor must have at least a start and end frame'))
        return warnings, errors

    if not actor.detections[0].frame <= actor.begin:
        errors.append(BoilerError(f"{prefix} actor's first detection must be <= actor begin frame"))
    if not (actor.detections[0].keyframe):
        errors.append(BoilerError(f'{prefix} first detection must be keyframe'))
    if not (actor.detections[-1].frame >= actor.end):
        errors.append(
            BoilerError(f"{prefix} actor's final detection must must be >= actor end frame")
        )
    if not (actor.detections[-1].keyframe):
        errors.append(BoilerError(f'{prefix} last detection must be keyframe'))

    total_frames = actor.detections[-1].frame - actor.detections[0].frame + 1
    if not (keyframe_count <= MAX_KEYFRAME_RATIO * total_frames):
        warnings.append(
            BoilerWarning(
                f'{prefix} keyframe density {round(keyframe_count / total_frames * 100)}%'
                f' higher than {round(MAX_KEYFRAME_RATIO * 100)}%'
            )
        )

    return warnings, errors


def validate_activity(activity: Activity) -> PROBLEMS:
    """
    * actors involved are appropriate for activity type
    * actors framerange is within activity framerange
    """
    warnings: List[BoilerWarning] = []
    errors: List[BoilerError] = []
    actor_string = ''
    activity_type = ActivityType(activity.activity_type)
    first_actor_frame = inf
    last_actor_frame = 0
    activity_id = activity.activity_id
    for actor in activity.actors:
        actor_warnings, actor_errors = validate_actor(actor, activity.activity_id, actor.actor_id)
        warnings += actor_warnings
        errors += actor_errors
        actor_type = actor.actor_type
        actor_string += actor_codes[actor_type]
        first_actor_frame = min(first_actor_frame, actor.begin)
        last_actor_frame = max(last_actor_frame, actor.end)
        if actor.begin > activity.end:
            errors.append(
                BoilerError(
                    f'activity_id={activity_id} actor track must begin before activity ends'
                )
            )
        if actor.end < activity.begin:
            errors.append(
                BoilerError(f'activity_id={activity_id} actor track must end after activity begins')
            )
    if first_actor_frame > activity.begin:
        errors.append(
            BoilerError(
                f'activity_id={activity_id} there must be an actor at the first frame of an activity'  # noqa:E501
            )
        )
    if last_actor_frame < activity.end:
        errors.append(
            BoilerError(
                f'activity_id={activity_id} there must be an actor at the last frame of an activity'
            )
        )
    at_warnings, at_errors = validate_activity_actor_types(actor_string, activity_type, activity_id)
    return warnings + at_warnings, errors + at_errors


def validate_activities(activity_list: ActivityList) -> PROBLEMS:
    warnings: List[BoilerWarning] = []
    errors: List[BoilerError] = []
    for activity in activity_list:
        activity_warnings, activity_errors = validate_activity(activity)
        warnings += activity_warnings
        errors += activity_errors
    return warnings, errors



from boiler.definitions import ActivityType

# This translates from the old activity names (prior to March 2020) to the new names.
activity_type_mapping = {
    'Abandon_Package': ActivityType.PERSON_ABANDONS_PACKAGE,
    'Carry_in_Hands': ActivityType.CARRY_IN_HANDS,
    'Carry_on_Back': ActivityType.CARRY_ON_BACK,
    'Casing_Facility': ActivityType.CASING_FACILITY,
    'Close_Facility_Door': ActivityType.PERSON_CLOSES_FACILITY_DOOR,
    'Close_Trunk': ActivityType.PERSON_CLOSES_TRUNK,
    'Close_Vehicle_Door': ActivityType.PERSON_CLOSES_VEHICLE_DOOR,
    'Embrace_Interaction': ActivityType.PERSON_EMBRACES_PERSON,
    'Enter_Facility': ActivityType.PERSON_ENTERS_SCENE_THROUGH_STRUCTURE,
    'Enter_Through_Structure': ActivityType.PERSON_ENTERS_SCENE_THROUGH_STRUCTURE,
    'Enter_Vehicle': ActivityType.PERSON_ENTERS_VEHICLE,
    'Exit_Facility': ActivityType.PERSON_EXITS_SCENE_THROUGH_STRUCTURE,
    'Exit_Through_Structure': ActivityType.PERSON_EXITS_SCENE_THROUGH_STRUCTURE,
    'Exit_Vehicle': ActivityType.PERSON_EXITS_VEHICLE,
    'Hand_Interaction': ActivityType.HAND_INTERACTS_WITH_PERSON,
    'Heavy_Carry': ActivityType.PERSON_CARRIES_HEAVY_OBJECT,
    'Joining_Queue': ActivityType.JOINING_QUEUE,
    'Laptop_Interaction': ActivityType.PERSON_INTERACTS_WITH_LAPTOP,
    'Load_Vehicle': ActivityType.PERSON_LOADS_VEHICLE,
    'Object_Transfer': ActivityType.PERSON_TRANSFERS_OBJECT,
    'Open_Facility_Door': ActivityType.PERSON_OPENS_FACILITY_DOOR,
    'Open_Trunk': ActivityType.PERSON_OPENS_TRUNK,
    'Open_Vehicle_Door': ActivityType.PERSON_OPENS_VEHICLE_DOOR,
    'People_Talking': ActivityType.PERSON_TALKS_TO_PERSON,
    'Pick_Up_Object': ActivityType.PERSON_PICKS_UP_OBJECT,
    'Purchasing': ActivityType.PERSON_PURCHASES,
    'Put_Down_Object': ActivityType.PERSON_PUTS_DOWN_OBJECT,
    'Read_Document': ActivityType.PERSON_READS_DOCUMENT,
    'Riding': ActivityType.PERSON_RIDES_BICYCLE,
    'Set_Down_Object': ActivityType.PERSON_PUTS_DOWN_OBJECT,
    'Sit_Down': ActivityType.PERSON_SITS_DOWN,
    'Stand_Up': ActivityType.PERSON_STANDS_UP,
    'Talk_on_Phone': ActivityType.PERSON_STANDS_UP,
    'Text_On_Phone': ActivityType.PERSON_TEXTS_ON_PHONE,
    'Theft': ActivityType.PERSON_STEALS_OBJECT,
    'Unload_Vehicle': ActivityType.PERSON_UNLOADS_VEHICLE,
    'Vehicle_DropsOff_Person': ActivityType.VEHICLE_DROPS_OFF_PERSON,
    'Vehicle_Moving': ActivityType.VEHICLE_MOVING,
    'Vehicle_PicksUp_Person': ActivityType.VEHICLE_PICKS_UP_PERSON,
    'Vehicle_Reversing': ActivityType.VEHICLE_REVERSES,
    'Vehicle_Starting': ActivityType.VEHICLE_STARTS,
    'Vehicle_Stopping': ActivityType.VEHICLE_STOPS,
    'Vehicle_Turning_Left': ActivityType.VEHICLE_TURNS_LEFT,
    'Vehicle_Turning_Right': ActivityType.VEHICLE_TURNS_RIGHT,
    'Vehicle_UTurn': ActivityType.VEHICLE_MAKES_U_TURN,
}



"""
Deserialize kpf to boiler_rest client types
"""
from io import BytesIO
import json
from typing import Any, Dict, IO, List, Optional
from uuid import UUID
from zipfile import ZipFile

import yaml

from boiler import BoilerError
from boiler.definitions import ActivityPipelineStatuses
from boiler.models import Activity, ActivityList, Actor, Box, Detection


GEOM_ID = 'id0'  # detection id
ACTOR_ID = 'id1'  # actor id
CSET3 = 'cset3'
GEOM = 'geom'
TYPES = 'types'
FRAME = 'ts0'  # timestamp in frames
SECONDS = 'ts1'  # timestamp in seconds
BOX = 'g0'  # geometry bounding box: [top, left, bottom, right]
KEYFRAME = 'keyframe'  # boolean is keyframe?
ACTIVITY = 'act'
ACTIVITY_TYPE = 'act2'
ACTIVITY_ID = 'id2'
TIMESPANS = 'timespan'
FRAME_TIMESPAN = 'tsr0'
STATUS = 'src_status'
ACTORS = 'actors'

_max_line_width = 1 << 30  # need a 32 bit signed int


def dump_yaml_line(content: Any) -> str:
    return f'- {json.dumps(content)}\n'


def load_yaml(file):
    try:
        return yaml.load(file, Loader=yaml.CLoader)
    except AttributeError:
        return yaml.safe_load(file)


# TODO: Fix kpf deserialization and remove type: ignore


def deserialize_types(file, actor_map):
    yml = load_yaml(file)
    for type_packet in yml:
        if TYPES in type_packet:
            kpf_types = type_packet[TYPES]
            id1 = kpf_types[ACTOR_ID]
            cset3 = kpf_types[CSET3]
            cset3keys = list(cset3.keys())
            if len(cset3keys) != 1:
                raise BoilerError(f'{CSET3} should only have 1 key, found {cset3keys}')
            name = cset3keys[0]
            if id1 in actor_map:
                actor_map[id1].actor_type = name
            else:
                actor_map[id1] = Actor(actor_type=name)  # type: ignore


def deserialize_geom(file, actor_map):
    """
    python's yaml parsing is painfully slow,
    this func could take on the order of 10 seconds
    """
    yml = load_yaml(file)
    for geom_packet in yml:
        if GEOM in geom_packet:
            geom = geom_packet[GEOM]
            actor_id = geom[ACTOR_ID]
            frame = geom[FRAME]
            box = geom[BOX]
            box = [int(n) for n in box.split(' ')]
            if len(box) != 4:
                raise BoilerError('expect bounding box to have 4 values')
            keyframe = False
            if KEYFRAME in geom:
                keyframe = geom[KEYFRAME]
            box = Box(left=box[0], top=box[1], right=box[2], bottom=box[3])
            detection = Detection(frame=frame, box=box, keyframe=keyframe)
            if actor_id in actor_map:
                actor_map[actor_id].detections.append(detection)
            else:
                actor_map[actor_id] = Actor(  # type: ignore
                    clip_id=actor_id, detections=[detection]
                )

    for actor in actor_map.values():
        actor.sort_detections()


def _deserialize_frame_timespan(timespans):
    for timespan in timespans:
        if FRAME_TIMESPAN in timespan:
            return timespan[FRAME_TIMESPAN]
    raise BoilerError(f'no timespan in {timespans} with {FRAME_TIMESPAN}')


def _deserialize_actor(actor, actor_map):
    if ACTOR_ID not in actor:
        raise BoilerError(f'actor {actor} missing {ACTOR_ID}')
    if TIMESPANS not in actor:
        raise BoilerError(f'actor {actor} missing {TIMESPANS}')
    actor_id = actor[ACTOR_ID]
    timespans = actor[TIMESPANS]
    frame_timespan = _deserialize_frame_timespan(timespans)
    if actor_id in actor_map:
        actor_map[actor_id].begin = frame_timespan[0]
        actor_map[actor_id].end = frame_timespan[1]
    else:
        actor_map[actor_id] = Actor(  # type: ignore
            clip_id=actor_id, begin=frame_timespan[0], end=frame_timespan[1]
        )
    return actor_map[actor_id]


def _deserialize_activity(activity_packet, actor_map):
    """
    returns activity instance
    """
    activity = activity_packet[ACTIVITY]
    timespans = activity[TIMESPANS]
    frame_timespan = _deserialize_frame_timespan(timespans)
    actors = activity[ACTORS]
    activity_type_obj = activity[ACTIVITY_TYPE]
    activity_type_keys = list(activity_type_obj.keys())
    if len(activity_type_keys) != 1:
        raise BoilerError(f'{activity_type_obj} should only have 1 key')
    activity_type = activity_type_keys[0]
    status = None
    if STATUS in activity:
        status = ActivityPipelineStatuses(activity['status'])
    return Activity(  # type: ignore
        clip_id=activity[ACTIVITY_ID],
        activity_type=activity_type,
        begin=frame_timespan[0],
        end=frame_timespan[1],
        status=status,
        actors=[_deserialize_actor(a, actor_map) for a in actors],
    )


def deserialize_activities(file, activity_map, actor_map):
    """
    Load activity from file given geometry and types
    returns map from activity ID to activity
    """
    yml = load_yaml(file)
    for activity_packet in yml:
        if ACTIVITY in activity_packet:
            activity = _deserialize_activity(activity_packet, actor_map)
            activity_map[activity.clip_id] = activity


def serialize_types(actor_map: Dict[int, Actor]):
    for actor_id, actor in actor_map.items():
        line = {'types': {ACTOR_ID: actor_id, CSET3: {actor.actor_type.value: 1.0}}}
        yield dump_yaml_line(line)


def serialize_geom(actor_map: Dict[int, Actor]):
    geom_id = 1
    for actor_id, a in actor_map.items():
        for d in a.interpolated_detections:
            if a.begin <= d.frame <= a.end:
                line = {
                    'geom': {
                        ACTOR_ID: actor_id,
                        GEOM_ID: geom_id,
                        FRAME: d.frame,
                        # TODO ts1 requires framerate
                        BOX: f'{d.box.left} {d.box.top} {d.box.right} {d.box.bottom}',
                        KEYFRAME: d.keyframe,
                    }
                }
                geom_id = geom_id + 1
                yield dump_yaml_line(line)


def serialize_activities(activity_map: Dict[int, Activity], actor_map: Dict[int, Actor]):
    for activity_id, activity in activity_map.items():
        activity = activity_map[activity_id]

        line = {
            ACTIVITY: {
                ACTIVITY_TYPE: {activity.activity_type.value: 1.0,},
                ACTIVITY_ID: activity_id,
                TIMESPANS: [{FRAME_TIMESPAN: [activity.begin, activity.end]}],
                ACTORS: [
                    {
                        ACTOR_ID: actor.actor_id,
                        TIMESPANS: [{FRAME_TIMESPAN: [actor.begin, actor.end]}],
                    }
                    for actor in activity.actors
                ],
            }
        }

        if activity.status:
            line[ACTIVITY][STATUS] = activity.status.value

        yield dump_yaml_line(line)


def serialize_to_files(
    output_basename: str, activity_list: ActivityList, keyframes_only=False,
):
    actor_map = activity_list.actor_map
    activity_map = activity_list.activity_map

    with open(output_basename + '.activities.yml', 'w', encoding='utf-8') as yamlfile:
        for line in serialize_activities(activity_map, actor_map):
            yamlfile.write(line)
    with open(output_basename + '.types.yml', 'w', encoding='utf-8') as yamlfile:
        for line in serialize_types(actor_map):
            yamlfile.write(line)
    with open(output_basename + '.geom.yml', 'w', encoding='utf-8') as yamlfile:
        for line in serialize_geom(actor_map):
            yamlfile.write(line)


def serialize_to_zip(
    base_name: str,
    activity_list: ActivityList,
    meta: Optional[List[Any]] = None,
    video_id: Optional[UUID] = None,
) -> IO[bytes]:
    actor_map = activity_list.actor_map
    activity_map = activity_list.activity_map
    meta = meta or []
    meta_lines = [dump_yaml_line(m) for m in meta]
    activity_meta_lines = []

    for activity_type, activities in activity_list.activity_types.items():
        m = {'meta': f'{len(list(activities))} {activity_type.value} instances'}
        activity_meta_lines.append(dump_yaml_line(m))

    buffer = BytesIO()
    with ZipFile(buffer, 'w') as zip:
        lines = (
            meta_lines + activity_meta_lines + list(serialize_activities(activity_map, actor_map))
        )
        zip.writestr(f'{base_name}.activities.yml', ''.join(lines))

        lines = meta_lines + list(serialize_types(actor_map))
        zip.writestr(f'{base_name}.types.yml', ''.join(lines))

        lines = meta_lines + list(serialize_geom(actor_map))
        zip.writestr(f'{base_name}.geom.yml', ''.join(lines))

        if video_id:
            zip.writestr('stumpf_video_id.txt', str(video_id))

    return buffer
